/**
 * Program      i2sExploration.cpp
 * Author       2021-07-14 Charles Geiser (https://www.dodeka.ch)
 * 
 * Purpose      With the program we can investigate how the two hardware divisors influence 
 *              the frequencies of the I2S system. With selectable bit patterns we can observe 
 *              the interaction of the wordselect signal with the I2S output signal.
 * 
 *              There are 10 samples defined with 32 bits each, which can be selected in 
 *              sequence with a push button. These 32 bits are divided into 16 bits for 
 *              the right and 16 bits for the left audio channel.The currently output
 *              pattern is displayed on the monitor in binary form.
 * 
 * Board        Wemos D1 R2 with ESP8266
 *
 * Wiring       Oscilloscope on D8 f_bitclk  ( bitclock = 32 times the frequency at D4)
 *                              D4 f_ws      ( word select )
 *                              RS data out  ( I2S output signal, shiftet out sample)
 *                              D3 Pushbutton to GND, selects output sample
 * 
 * Remarks      The I2S bitclock and wordselect signals are derived from an internal base frequency
 *              I2SBASEFREQ of 160MHz which is divided by 32 to get f_baseclk of 5MHz.
 * 
 *              The WS signal transitions one clock period before the completion of a data word 
 *              The high word (16 bit) of a data double word (32 bit) represents the right channel and
 *              is transmitted on WS = 1 with MSB first. The low word is transmitted on WS = 0
 * 
 * Formulas     f_baseclk = I2SBASEFREQ / 32                ( --> 5MHz )
 *              f_ws      = f_baseclk / (div1 * div2)       ( with div1, div2 in the range 2..63 )
 *              f_bitclk  = 32 * f_ws                       ( valid only when div1, div2 > 1)
 * 
 *              fws_min = 1259.76 Hz ≙ 7938 us (with div1 = div2 = 63) 
 *              fws_max = 1.25 MHz   ≙ 0.80 us (with div1 = div2 = 2)
 * 
 *               In order to set f_ws, we need to find two integer divisors whose product 
 *               approximates the value f_baseclk/f_ws as closely as possible (see function findDivisors() ).
 * 
 * Library      i2s_begin()
 * functions    i2s_set_rate()
 * used         i2s_set_dividers()
 *              i2s_get_real_rate()
 *              i2s_write_sample()
 *              
 * References   https://www.allaboutcircuits.com/technical-articles/introduction-to-the-i2s-interface/
 *              https://github.com/esp8266/Arduino/pull/4031
 *              https://github.com/espressif/ESP8266_MP3_DECODER/issues/3
 *              https://github.com/esp8266/Arduino/pull/4031
 *              https://shepherdingelectrons.blogspot.com/2018/09/esp8266-minimum-i2s-code.html
 *              https://www.youtube.com/watch?v=p6BzdsJR4mE
 */
#include <Arduino.h>
#include <core_esp8266_i2s.h>
#include "ButtonDebounced.h"

#define pinButton D3

// Forward declaration of functions
void nextSample();
void prevSample();
void showCurrent();

ButtonDebounced button(pinButton, nextSample, prevSample, showCurrent);
//                                click       longclick   doubleclick

typedef struct { const char *name; const char *pat; uint32_t value; } Bitpattern;
Bitpattern sample[] =
{
  //                                             |   right chn  ||   left chn   |
  { "0", "0b10000000000000001010000000000000", 0b10000000000000001010000000000000 },
  { "1", "0b11111111111111110000000000000000", 0b11111111111111110000000000000000 },
  { "2", "0b11111111000000001111111100000000", 0b11111111000000001111111100000000 },
  { "3", "0b11110000111100001111000011110000", 0b11110000111100001111000011110000 },
  { "4", "0b11001100110011001100110011001100", 0b11001100110011001100110011001100 },
  { "5", "0b10101010101010101010101010101010", 0b10101010101010101010101010101010 },
  { "6", "0b00000000000000011000000000000000", 0b00000000000000011000000000000000 },
  { "7", "0b00000000000000111100000000000000", 0b00000000000000111100000000000000 },
  { "8", "0b11111111111111000011111111111111", 0b11111111111111000011111111111111 },
  { "9", "0b11111111111111100111111111111111", 0b11111111111111100111111111111111 },
  
};
constexpr uint8_t nbrSamples = sizeof(sample) / sizeof(sample[0]);

uint32_t msPrevious = millis();
int div1;
int div2;
int zbest;
float diffbest;
constexpr float f_baseclk = I2SBASEFREQ / 32;

// Set your sampling rate on line below
//constexpr float f_ws = 44100;        // = f_bck / (57 x 2)
//constexpr float f_ws = 48000;        // f_bck / (52 x 2)
constexpr float f_ws = 2500;         // = f_bck / (50 x 40)

constexpr float div1xdiv2 = f_baseclk / f_ws; // div1xdiv2 = 1..63 x 1..63 = 1..3969

// What happens when div1 or div2 are less than 2 ? 
//  div1  div2  f_ws [MHz]  f_bitclk [MHz]  f_bitclk/f_ws   f_ws ?= f_baseclk/div1xdiv2
//    1     1     2.5         40              16              no
//    1     2     2.5         40              16              no
//    2     1     1.25        40              32              no
//    2     2     1.25        40              32              yes
//    1     3     1.666       26.66           16              yes
//    3     1     0.833       26.66           32              no
//    2     3     0.833       26.66           32              yes
//    3     2     0.833       26.66           32              yes

uint8_t currentSample = 0;
const uint8_t pinLED = LED_BUILTIN;  // GPIO2
const uint8_t pinI2SO_DATA = RX;     // GPIO3
const uint8_t pinI2SO_BCK  = D8;     // GPIO15
const uint8_t pinI2SO_WS   = D4;     // GPIO2

/**
 * Select next sample on button click
 */
void nextSample()
{
  currentSample++;
  if (currentSample == nbrSamples) currentSample = 0;
  showCurrent();
}

/**
 * Select previous sample on button longclick
 */
void prevSample()
{
  if (currentSample == 0) currentSample = nbrSamples;
  currentSample--;
  showCurrent();
}

/**
 * Print name and value of current sample
 */
void showCurrent() 
{
  Serial.printf_P(PSTR("Current Sample: %s --> %s\n"), sample[currentSample].name, sample[currentSample].pat);
}; 

/**
 * Finds the best approximation for z = div1 * div2 when z is a 
 * floating point number not greater than 63 * 63 = 3969 and the 
 * divisors are integers in the range 2..63
 */
void findDivisors(float z, int &d1, int &d2, int &zb, float &db)
{
  int max_x = round(sqrt(z)) + 1;
  int max_y = 64;
  int i = 0;
  float diff;
  db = max_y * max_y;
  for (int x = 2; x < max_x; x++)
  {
    for (int y = x; y < max_y; y++)
    {
      diff = abs(z - x * y);
      i++;
      if (diff < db)
      {
        diffbest = diff;
        d1 = x;
        d2 = y;
        zb = x * y;
        if (d1 < d2) { d1 = y; d2 = x; }
        //Serial.printf("i=%d, x=%d, y=%d, zbest=%d, diff=%f,\n", i, d1, d2, zb, db);
      }
    }
  }
}

void showValues()
{
  Serial.printf("target f_ws = %.3f, target div1 x div2 = %.3f, real f_ws = %.3f\n", f_ws, f_baseclk/f_ws, i2s_get_real_rate());
  Serial.printf("div1 = %d, div2 = %d, div1 x div2 = %d\n\n", div1, div2, div1 * div2);
}

/**
 * Returns true, as soon as msWait milliseconds have passed.
 * Supply a reference to an unsigned long variable to hold 
 * the previous milliseconds.
 */
bool waitIsOver(uint32_t &msPrevious, uint32_t msWait)
{
  return (millis() - msPrevious >= msWait) ? (msPrevious = millis(), true) : false;
}

void setup()
{
  Serial.begin(115200);
  i2s_begin();
  i2s_set_rate(f_ws);
  //i2s_set_dividers(40, 50); // is equivalent to line above when f_ws = 2500
}

bool done = false;
void loop()
{ 
  // delay(10000); return; // activate this line to investigate only the set frequencies
  if(!done)
  {
    Serial.println(); 
    Serial.println();
    findDivisors(div1xdiv2, div1, div2, zbest, diffbest);
    showValues();
    done = true;
  }

  i2s_write_sample(sample[currentSample].value); // write a doube word
  button.loop();
}